UPSTREAM: mm: reuse only-pte-mapped KSM page in do_wp_page()
authorKirill Tkhai <ktkhai@virtuozzo.com>
Tue, 5 Mar 2019 23:43:06 +0000 (15:43 -0800)
committerTodd Kjos <tkjos@google.com>
Thu, 7 Jul 2022 02:21:25 +0000 (19:21 -0700)
commit 52d1e606ee733921e984770d47539a6bb91e8506 upstream.

Add an optimization for KSM pages almost in the same way that we have
for ordinary anonymous pages.  If there is a write fault in a page,
which is mapped to an only pte, and it is not related to swap cache; the
page may be reused without copying its content.

[ Note that we do not consider PageSwapCache() pages at least for now,
  since we don't want to complicate __get_ksm_page(), which has nice
  optimization based on this (for the migration case). Currenly it is
  spinning on PageSwapCache() pages, waiting for when they have
  unfreezed counters (i.e., for the migration finish). But we don't want
  to make it also spinning on swap cache pages, which we try to reuse,
  since there is not a very high probability to reuse them. So, for now
  we do not consider PageSwapCache() pages at all. ]

So in reuse_ksm_page() we check for 1) PageSwapCache() and 2)
page_stable_node(), to skip a page, which KSM is currently trying to
link to stable tree.  Then we do page_ref_freeze() to prohibit KSM to
merge one more page into the page, we are reusing.  After that, nobody
can refer to the reusing page: KSM skips !PageSwapCache() pages with
zero refcount; and the protection against of all other participants is
the same as for reused ordinary anon pages pte lock, page lock and
mmap_sem.

[akpm@linux-foundation.org: replace BUG_ON()s with WARN_ON()s]
Link: http://lkml.kernel.org/r/154471491016.31352.1168978849911555609.stgit@localhost.localdomain
Signed-off-by: Kirill Tkhai <ktkhai@virtuozzo.com>
Reviewed-by: Yang Shi <yang.shi@linux.alibaba.com>
Cc: "Kirill A. Shutemov" <kirill@shutemov.name>
Cc: Hugh Dickins <hughd@google.com>
Cc: Andrea Arcangeli <aarcange@redhat.com>
Cc: Christian Koenig <christian.koenig@amd.com>
Cc: Claudio Imbrenda <imbrenda@linux.vnet.ibm.com>
Cc: Rik van Riel <riel@surriel.com>
Cc: Huang Ying <ying.huang@intel.com>
Cc: Minchan Kim <minchan@kernel.org>
Cc: Kirill Tkhai <ktkhai@virtuozzo.com>
Signed-off-by: Andrew Morton <akpm@linux-foundation.org>
Signed-off-by: Linus Torvalds <torvalds@linux-foundation.org>
Change-Id: Ic301afb9d0739de5643209a4c753728a9e20d411

include/linux/ksm.h
mm/ksm.c
mm/memory.c

index 44368b19b27e11e848ced6348c4e747396c23b6b..def48a2d87aa90f28bcdc23ce32b0b32ab7852ee 100644 (file)
@@ -64,6 +64,8 @@ struct page *ksm_might_need_to_copy(struct page *page,
 
 void rmap_walk_ksm(struct page *page, struct rmap_walk_control *rwc);
 void ksm_migrate_page(struct page *newpage, struct page *oldpage);
+bool reuse_ksm_page(struct page *page,
+                       struct vm_area_struct *vma, unsigned long address);
 
 #else  /* !CONFIG_KSM */
 
@@ -103,6 +105,11 @@ static inline void rmap_walk_ksm(struct page *page,
 static inline void ksm_migrate_page(struct page *newpage, struct page *oldpage)
 {
 }
+static inline bool reuse_ksm_page(struct page *page,
+                       struct vm_area_struct *vma, unsigned long address)
+{
+       return false;
+}
 #endif /* CONFIG_MMU */
 #endif /* !CONFIG_KSM */
 
index 6d3bc2723c9b3d8ce548593a7963cd68342ff31d..60505a0c1142b0d00390f9ca8ebab35bc3691602 100644 (file)
--- a/mm/ksm.c
+++ b/mm/ksm.c
@@ -695,8 +695,9 @@ again:
         * case this node is no longer referenced, and should be freed;
         * however, it might mean that the page is under page_freeze_refs().
         * The __remove_mapping() case is easy, again the node is now stale;
* but if page is swapcache in migrate_page_move_mapping(), it might
* still be our page, in which case it's essential to keep the node.
         */
        while (!get_page_unless_zero(page)) {
                /*
@@ -2610,6 +2611,31 @@ again:
                goto again;
 }
 
+bool reuse_ksm_page(struct page *page,
+                   struct vm_area_struct *vma,
+                   unsigned long address)
+{
+#ifdef CONFIG_DEBUG_VM
+       if (WARN_ON(is_zero_pfn(page_to_pfn(page))) ||
+                       WARN_ON(!page_mapped(page)) ||
+                       WARN_ON(!PageLocked(page))) {
+               dump_page(page, "reuse_ksm_page");
+               return false;
+       }
+#endif
+
+       if (PageSwapCache(page) || !page_stable_node(page))
+               return false;
+       /* Prohibit parallel get_ksm_page() */
+       if (!page_ref_freeze(page, 1))
+               return false;
+
+       page_move_anon_rmap(page, vma);
+       page->index = linear_page_index(vma, address);
+       page_ref_unfreeze(page, 1);
+
+       return true;
+}
 #ifdef CONFIG_MIGRATION
 void ksm_migrate_page(struct page *newpage, struct page *oldpage)
 {
index 4154fb45ac0f7c9e14c82184d29fa1a40b010968..08a525aeec0e5b1bd313f1bcf8fbe59fd6b6d1f4 100644 (file)
@@ -2854,8 +2854,11 @@ static int do_wp_page(struct vm_fault *vmf)
         * Take out anonymous pages first, anonymous shared vmas are
         * not dirty accountable.
         */
-       if (PageAnon(vmf->page) && !PageKsm(vmf->page)) {
+       if (PageAnon(vmf->page)) {
                int total_map_swapcount;
+               if (PageKsm(vmf->page) && (PageSwapCache(vmf->page) ||
+                                          page_count(vmf->page) != 1))
+                       goto copy;
                if (!trylock_page(vmf->page)) {
                        get_page(vmf->page);
                        pte_unmap_unlock(vmf->pte, vmf->ptl);
@@ -2870,6 +2873,15 @@ static int do_wp_page(struct vm_fault *vmf)
                        }
                        put_page(vmf->page);
                }
+               if (PageKsm(vmf->page)) {
+                       bool reused = reuse_ksm_page(vmf->page, vmf->vma,
+                                                    vmf->address);
+                       unlock_page(vmf->page);
+                       if (!reused)
+                               goto copy;
+                       wp_page_reuse(vmf);
+                       return VM_FAULT_WRITE;
+               }
                if (reuse_swap_page(vmf->page, &total_map_swapcount)) {
                        if (total_map_swapcount == 1) {
                                /*
@@ -2890,7 +2902,7 @@ static int do_wp_page(struct vm_fault *vmf)
                                        (VM_WRITE|VM_SHARED))) {
                return wp_page_shared(vmf);
        }
-
+copy:
        /*
         * Ok, we need to copy. Oh, well..
         */
@@ -695,8 +695,9 @@ again:
         * however, it might mean that the page is under page_freeze_refs().
         * The __remove_mapping() case is easy, again the node is now stale;
-        * but if page is swapcache in migrate_page_move_mapping(), it might
-        * still be our page, in which case it's essential to keep the node.
+        * the same is in reuse_ksm_page() case; but if page is swapcache
+        * in migrate_page_move_mapping(), it might still be our page,
+        * in which case it's essential to keep the node.
         */
        while (!get_page_unless_zero(page)) {